// Copyright 2010
// Author: Li Zhang

#include <gtest/gtest.h>
#include <stdio.h>
#include <stdint.h>
#include <iostream>
#include <string>
#include <vector>
#include <utility>
#include "./ZukerAlgorithm.h"
using std::cout;
using std::endl;
#define a -3.0
#define INF 100

// Method for calculating the energies of Stacking loop.
float ZukerAlgorithm::_eS(char x, char y)
{
  if ((x == 'g' && y == 'c') || (x == 'c' && y == 'g'))
    return -12.0;
  else if ((x == 'a' && y == 'u') || (x == 'u' && y == 'a'))
    return -12.1;
  else if ((x == 'g' && y == 'u') || (x == 'u' && y == 'g'))
    return -11.8;
  else
    return INF;
}

// Method for calculating the energies of Hairpin loop.
float ZukerAlgorithm::_eH(char x, char y)
{
  if ((x == 'g' && y == 'c') || (x == 'c' && y == 'g'))
    return -20.4;
  else if ((x == 'a' && y == 'u') || (x == 'u' && y == 'a'))
    return -20.7;
  else if ((x == 'g' && y == 'u') || (x == 'u' && y == 'g'))
    return -20.5;
  else
    return INF;
}

// Method for calculating the energies of bulge or interior loop.
float ZukerAlgorithm::_eBI(char i, char j, char x, char y)
{
  if ((i == 'a' && j == 'u' && x == 'g' && y == 'c') ||
  (i == 'u' && j == 'a' && x == 'c' && y == 'g') ||
  (i == 'a' && j == 'u' && x == 'c' && y == 'g') ||
  (i == 'u' && j == 'a' && x == 'g' && y == 'c'))
    return -8.1;
  else if ((i == 'a' && j == 'u' && x == 'u' && y == 'g') ||
  (i == 'a' && j == 'u' && x == 'g' && y == 'u') ||
  (i == 'u' && j == 'a' && x == 'u' && y == 'g') ||
  (i == 'u' && j == 'a' && x == 'g' && y == 'u'))
    return -8.2;
  else if (( i == 'a' && j == 'u' && x == 'a' && y == 'u') ||
  (i == 'a' && j == 'u' && x == 'u' && y == 'a') ||
  (i == 'u' && j == 'a' && x == 'a' && y == 'u') ||
  (i == 'u' && j == 'a' && x == 'u' && y == 'a'))
    return -8.3;
  else if ((i == 'c' && j == 'g' && x == 'c' && y == 'g') ||
  (i == 'c' && j == 'g' && x == 'g' && y == 'c') ||
  (i == 'g' && j == 'c' && x == 'c' && y == 'g') ||
  (i == 'g' && j == 'c' && x == 'g' && y == 'c'))
    return -8.15;
  else if ((i == 'c' && j == 'g' && x == 'a' && y == 'u') ||
  (i == 'c' && j == 'g' && x == 'u' && y == 'a') ||
  (i == 'g' && j == 'c' && x == 'a' && y == 'u') ||
  (i == 'g' && j == 'c' && x == 'u' && y == 'a'))
    return -8.25;
  else if ((i == 'c' && j == 'g' && x == 'g' && y == 'u') ||
  (i == 'c' && j == 'g' && x == 'u' && y == 'g') ||
  (i == 'g' && j == 'c' && x == 'u' && y == 'g') ||
  (i == 'g' && j == 'c' && x == 'g' && y == 'u'))
    return -8.35;
  else if ((i == 'g' && j == 'u' && x == 'c' && y == 'g') ||
  (i == 'g' && j == 'u' && x == 'g' && y == 'c') ||
  (i == 'u' && j == 'g' && x == 'c' && y == 'g') ||
  (i == 'u' && j == 'g' && x == 'g' && y == 'c'))
    return -8.05;
  else if ((i == 'g' && j == 'u' && x == 'a' && y == 'u') ||
  (i == 'g' && j == 'u' && x == 'u' && y == 'a') ||
  (i == 'u' && j == 'g' && x == 'a' && y == 'u') ||
  (i == 'u' && j == 'g' && x == 'u' && y == 'a'))
    return -8.06;
  else if ((i == 'g' && j == 'u' && x == 'g' && y == 'u') ||
  (i == 'g' && j == 'u' && x == 'u' && y == 'g') ||
  (i == 'u' && j == 'g' && x == 'u' && y == 'g') ||
  (i == 'u' && j == 'g' && x == 'g' && y == 'u'))
    return -8.07;
  else
    return INF;
}

// Initialize the matrices.
void ZukerAlgorithm::initZukerAlgorithm(const std::string& input)
{
  _seqString = input;
  // Matrix _W.
  _W.resize(_seqString.length());
  for (int i = 0; i < static_cast<int>(_seqString.length()); ++i)
  {
    _W[i].resize(_seqString.length());
  }
  for (int i = 0; i < static_cast<int>(_seqString.length()); ++i)
  {
    for (int j = 0; j <= i; ++j)
    {
      _W[i][j] = 0;
    }
  }
  // Matrix _V.
  _V.resize(_seqString.length());
  for (int i = 0; i < static_cast<int>(_seqString.length()); ++i)
  {
    _V[i].resize(_seqString.length());
  }
  for (int i = 0; i < static_cast<int>(_seqString.length()); ++i)
  {
    for (int j = 0; j <= i; ++j)
    {
      _V[i][j] = INF;
    }
  }
  // Matrix _traceBackW.
  _traceBackW.resize(_seqString.length());
  for (int i = 0; i < static_cast<int>(_seqString.length()); ++i)
  {
    _traceBackW[i].resize(_seqString.length());
  }
  for (int i = 0; i < static_cast<int>(_seqString.length()); ++i)
  {
    for (int j = 0; j < static_cast<int>(_seqString.length()); ++j)
    {
      _traceBackW[i][j] = 0;
    }
  }
  // Matrix _traceBackV.
  _traceBackV.resize(_seqString.length());
  for (int i = 0; i < static_cast<int>(_seqString.length()); ++i)
  {
    _traceBackV[i].resize(_seqString.length());
  }
  for (int i = 0; i < static_cast<int>(_seqString.length()); ++i)
  {
    for (int j = 0; j < static_cast<int>(_seqString.length()); ++j)
    {
      _traceBackV[i][j] = 0;
    }
  }
}

// ___________________________________________________________________________
// Calculate the values of entries of matrix _W, which is the minimum energy
// matrix.
void ZukerAlgorithm::computeW()
{
  int i, j, k;
  float val, min = 0;
  for (int d = 1; d < static_cast<int>(_seqString.length()); d++)
  {
    for (i = 0; i < static_cast<int>(_seqString.length()); i++)
    {
      j = i + d;
      if (j < static_cast<int>(_seqString.length()))
      {
        min = 0;
        // Case 1.
        if (_W[i][j - 1] < min)
        {
          min = _W[i][j - 1];
          _traceBackW[i][j] = -1;
        }
        if (_W[i + 1][j] < min)
        {
          min = _W[i + 1][j];
          _traceBackW[i][j] = -2;
        }
        // Case 2.
        if (computeV(i, j) < min)
        {
          min = computeV(i, j);
          _traceBackW[i][j] = -3;
        }
        // Case 3.
        for (k = i + 1; k < j; k++)
        {
          val = _W[i][k] + _W[k + 1][j];
          if (val < min)
          {
            min = val;
            _traceBackW[i][j] = k;
          }
        }
        _W[i][j] = min;
      }
    }
  }
}

// Calculate the minimum energy of matrix _V.
float ZukerAlgorithm:: computeV(int i, int j)
{
  float min = 0, tmp = 0;
  // Case 1.
  if (_eH(_seqString.at(i), _seqString.at(j)) < min)
  {
    min = _eH(_seqString.at(i), _seqString.at(j));
    _traceBackV[i][j] = -1;
  }
  // Case 2.
  if (_eS(_seqString.at(i), _seqString.at(j)) + _V[i + 1][j - 1] < min)
  {
    min = _eS(_seqString.at(i), _seqString.at(j)) + _V[i + 1][j - 1];
    _traceBackV[i][j] = -2;
  }
  // Case 3.
  for (int k = i + 1; k < j - 1; ++k)
  {
    tmp = _W[i + 1][k] + _W[k + 1][j - 1] + a;
    if (tmp < min)
    {
      min = tmp;
      _traceBackV[i][j] = k;
    }
  }
  // Case 4.
  for (int k = i + 1; k < j - 2; ++k)
  {
    for (int l = i + 2; l < j - 1; ++l)
    {
      if ((k - i + j - l) > 2)
      tmp = _V[k][l] + _eBI(_seqString.at(i), _seqString.at(j),
          _seqString.at(k), _seqString.at(l));
      if (tmp < min)
      {
        min = tmp;
        _traceBackV[i][j] = -3;
      }
    }
  }
  _V[i][j] = min;
  return _V[i][j];
}


// Traceback to get the corresponding best structure.
void ZukerAlgorithm::traceBackW(int i, int j,
std::vector<std::pair<int, int> >* result)
{
  if (_traceBackW[i][j] == 0)
  {
    return;
  }
  else if (_traceBackW[i][j] == -1)
  {
    traceBackW(i, j - 1, result);
  }
  else if (_traceBackW[i][j] == -2)
  {
    traceBackW(i + 1, j, result);
  }
  else if (_traceBackW[i][j] == -3)
  {
    traceBackV(i, j, result);
  }
  else
  {
    int k = _traceBackW[i][j];
    if ((i + 1 <= k) &&(k + 1 <= j))
    {
      traceBackW(i, k, result);
      traceBackW(k + 1, j, result);
    }
  }
}


// Traceback to get the corresponding best structure.
void ZukerAlgorithm::traceBackV(int i, int j,
    std::vector<std::pair<int, int> >* result)
{
  if (_traceBackV[i][j] == 0)
  {
    return;
  }
  else if (_traceBackV[i][j] == -1)
  {
    result -> push_back(std::make_pair(i, j));
  }
  else if (_traceBackV[i][j] == -2)
  {
    traceBackV(i + 1, j - 1, result);
    result -> push_back(std::make_pair(i, j));
  }
  else if (_traceBackV[i][j] == -3)
  {
    result -> push_back(std::make_pair(i, j));
    traceBackV(i + 1, j - 1, result);
  }
  else
  {
    int k = _traceBackV[i][j];
    if ((i + 1 <= k) && (k + 1 <= j - 1))
    {
      traceBackW(i + 1, k, result);
      traceBackW(k + 1, j - 1, result);
    }
  }
}



// Compute the structure of the input sequence.
void ZukerAlgorithm::computeStructure(const std::string& input,
std::vector<std::pair<int, int> >* result)
{
  initZukerAlgorithm(input);
  computeW();
  traceBackW(0, _seqString.length() - 1, result);
}


// ___________________________________________________________________________
// Print the matrices.
void ZukerAlgorithm::printMatrix() const
{
  printf("\nTHIS IS MATRIX  _W:\n");
  for (size_t i = 0; i < _seqString.length(); ++i)
  {
    for (size_t j = 0; j < _seqString.length(); ++j)
    {
      cout << " " << _W[i][j];
    }
    cout << endl;
  }
  printf("\nTHIS IS MATRIX _V:\n");
  for (size_t i = 0; i < _seqString.length(); ++i)
  {
    for (size_t j = 0; j < _seqString.length(); ++j)
    {
      cout << " " << _V[i][j];
    }
    cout << endl;
  }
  printf("\nTHIS IS MATRIX _TRACEBACKW:\n");
  for (size_t i = 0; i < _seqString.length(); ++i)
  {
    for (size_t j = 0; j < _seqString.length(); ++j)
    {
      cout << " " << _traceBackW[i][j];
    }
    cout << endl;
  }
  printf("\nTHIS IS MATRIX _TRACEBACKV:\n");
  for (size_t i = 0; i < _seqString.length(); ++i)
  {
    for (size_t j = 0; j < _seqString.length(); ++j)
    {
      cout << " " << _traceBackV[i][j];
    }
    cout << endl;
  }
}

// Output the corresponding structure of the given string.
void ZukerAlgorithm::printStructure(const std::vector<std::pair<int, int> >&
    result) const
{
  printf("The CORRESPONDING RNA STRUCTURE IS:\n");
  for (size_t i = 0; i < result.size(); i++)
  {
    printf("(%u,%u)", result[i].first + 1, result[i].second + 1);
  }
  printf("\n");
}

// Method for calculating the running time and writing to a file.
void ZukerAlgorithm::performance(const std::string& s)
{
  const std::string inputString = s;
  FILE *fp;
  int flag = 1;
  float time_use = 0;
  struct timeval start;
  struct timeval end;
  fp = fopen("ZukerAlgorithmPerf.output.txt", "a");
  if (fp == NULL)
  {
    fprintf(stderr, "Could not open output file ZukerAlgorithmPerf.output.txt");
    flag = 0;
    exit(1);
  }
  if (flag)
  {
    std::vector<std::pair<int, int> > result;
    gettimeofday(&start, NULL);
    computeStructure(inputString, &result);
    gettimeofday(&end, NULL);
    time_use = (end.tv_sec - start.tv_sec) * 1000000 +
    (end.tv_usec - start.tv_usec);
    _dictionary.push_back(time_use);
    {
      fprintf(fp, "%d\t%f\n", s.length(), time_use);
    }
  }
  fclose(fp);
}

